import hljs from "highlight.js";
import cheerio from "cheerio";
import camelCase from "camelcase";
import chalk from "chalk";
import stripAnsi from "strip-ansi";
import merge from "deepmerge";
import ansiMark from "@central419/ansi-mark";
import stripIndent from "strip-indent";
import detectIndent from "detect-indent";

const darkPalette = {
	addition: chalk.green,
	attr: chalk.yellow,
	attribute: chalk.blue,
	attrString: chalk.cyan,
	base: chalk.white,
	builtIn: chalk.blue,
	builtInName: chalk.blue,
	bullet: chalk.magenta,
	class: chalk.green,
	code: chalk.yellow,
	comment: chalk.white.dim,
	deletion: chalk.red,
	doctag: chalk.blue,
	emphasis: chalk.magenta,
	function: chalk.white,
	formula: chalk.green,
	keyword: chalk.red,
	lineNumbers: chalk.grey,
	literal: chalk.magenta,
	link: chalk.blue.underline,
	meta: chalk.cyan,
	name: chalk.cyan,
	number: chalk.green,
	params: chalk.blue,
	quote: chalk.gray,
	regexp: chalk.magenta,
	selectorAttr: chalk.green,
	selectorClass: chalk.yellow,
	selectorId: chalk.blue,
	selectorPseudo: chalk.cyan,
	selectorTag: chalk.magenta,
	string: chalk.yellow,
	strong: chalk.red,
	subst: chalk.cyan,
	symbol: chalk.cyan,
	tag: chalk.blue,
	templateTag: chalk.magenta,
	templateVariable: chalk.green,
	title: chalk.green,
	trailingSpace: chalk,
	type: chalk.magenta,
	variable: chalk.red,
};

const filter = (node, options) => {
	let color;
	let text;
	let childText;

	if (node.type === "text") {
		text = node.data;
		return text;
	}

	if (node.name === "span" && node.type === "tag") {
		color = camelCase(node.attribs.class.split("-")[1]);
	}

	if (node.childNodes && node.childNodes.length > 0) {
		childText = node.childNodes
			.map((childNode) => filter(childNode, options))
			.join("");

		if (typeof color === "string" && options.colors[color]) {
			return options.colors[color](childText);
		}

		return childText;
	}

	return "";
};

const findLongestLine = (text, options) => {
	let tabPad = "";

	if (options.$indent.tabs) {
		tabPad = String().padEnd(options.consoleTabWidth, " ");
	}

	const lines = stripAnsi(text).replace(/\t/g, tabPad).split("\n");

	let max = 0;

	for (const line of lines) {
		if (line.length > max) {
			max = line.length;
		}
	}

	return max;
};

const padLine = (line, padding) => {
	const padString = String().padStart(padding, " ");
	return padString + line + padString;
};

const getIndentString = (options) => {
	if (options.$indent.tabs) {
		return String().padStart(1, "\t");
	}

	// Opts.$indent.spaces === true
	if (options.tabsToSpaces === 0) {
		return "\u0000";
	}

	return String().padEnd(options.tabsToSpaces, " ");
};

const syntaxHlString = (lang, script, options, indentStart) => {
	let output = "";

	const indentString = getIndentString(options);

	if (options.$indent.tabs) {
		script = script.replace(/\t/g, indentString);
	}

	if (options.$indent.spaces) {
		script = script.replace(/\t/g, indentString);
	}

	if (indentStart) {
		script = indentString + script;
	}

	const code = hljs.highlight(script, { language: lang }).value;
	const lines = code.split("\n");

	for (const [lineNumber, line] of lines.entries()) {
		const html = `<code>${line}</code>`;
		const $body = cheerio.load(html).root().find("code")[0];
		const text = filter($body, options);
		output += text;
		if (lineNumber !== lines.length - 1) {
			output += "\n";
		}
	}

	return output;
};

const syntaxHlJson = (json, options) => {
	const indentString = getIndentString(options);

	try {
		json = JSON.stringify(
			json,
			(key, value) => {
				if (value instanceof Function) {
					return `[FUNCTION]${String(value)}[FUNCTION]`;
				}

				return value;
			},
			indentString
		);
	} catch (error) {
		error.message = "🦅  Chromafi: " + error.message;
		throw new Error(error);
	}

	const highlighted = json.replace(
		/("(\\u[a-zA-Z\d]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?)/g,
		(match) => {
			let colorClass = "number";

			if (match.startsWith('"')) {
				if (match.endsWith(":")) {
					if (match.includes("-")) {
						colorClass = "attrString";
						match = match.replace(/"/g, "'");
					} else {
						colorClass = "attr";
						match = match.replace(/"/g, "");
					}
				} else {
					colorClass = "string";

					// eslint-disable-next-line unicorn/prefer-string-slice
					if (
						match.slice(1, 11) === "[FUNCTION]" &&
						match.substr(match.length - 11, 10) === "[FUNCTION]"
					) {
						colorClass = "function";
					} else {
						match = match.replace(/"/g, "'");
						match = match.replace(/\\n/g, "\n");
						match = match.replace(/\\t/g, indentString);
					}
				}
			} else if (/true|false/.test(match)) {
				colorClass = "literal";
			} else if (/null/.test(match)) {
				colorClass = "literal";
			}

			return options.colors[colorClass](match);
		}
	);

	const getFnStringIndent = (fnString, options_) => {
		fnString = fnString.replace(/\t/g, indentString);
		const indent = detectIndent(fnString);

		if (options_.$indent.spaces) {
			const indentLevel = indent.amount / options_.tabsToSpaces;
			return indentLevel;
		}

		if (options_.$indent.tabs) {
			const indentLevel = indent.amount;
			return indentLevel;
		}
	};

	const lines = highlighted
		.split("\n")
		.map((line) => {
			const fnParts = line.split("[FUNCTION]");
			if (fnParts.length === 3) {
				const plain = stripAnsi(line);

				let outerIndent;

				if (options.$indent.spaces) {
					outerIndent =
						plain.match(/^[ \\t]*/)[0].length /
						options.tabsToSpaces;
				}

				if (options.$indent.tabs) {
					outerIndent = plain.match(/^\t*/)[0].length;
				}

				const fnString = fnParts[1]
					.replace(/"/g, "'")
					.replace(/\\n/g, "\n")
					.replace(/\\t/g, "\t");

				const innerIndent = getFnStringIndent(fnString, options);
				const indentOffset = Math.abs(outerIndent - (innerIndent - 1));

				const re = new RegExp(`\n(\t){${indentOffset}}`, "g");
				const reTabbed = fnString
					.replace(re, "\n")
					.replace(/\\t/g, indentString);

				const preFn = fnParts[0].slice(
					0,
					Math.max(0, fnParts[0].length - 1)
				);
				const postFn = fnParts[2].slice(1);
				const jsHighlighted = syntaxHlString(
					"javascript",
					reTabbed,
					options
				);

				return preFn + jsHighlighted + postFn;
			}

			return line;
		})
		.join("\n");

	return lines;
};

const lineNumberPad = (number, options) => {
	if (!options.lineNumbers) {
		return "";
	}

	let output = "";

	const offsetLineN = number + (options.lineNumberStart - 1);

	if (options.$indent.spaces) {
		const padString = String().padStart(options.lineNumberPad, " ");
		const prePad = options.lineNumberPad + options.$maxDigitWidth;
		output = String(offsetLineN).padStart(prePad) + padString;
	}

	// Indent using spaces - up to the tabwidth required to contain number str
	if (options.$indent.tabs) {
		output = String(offsetLineN).padStart(options.$maxTabSpace, " ");
	}

	return options.colors.lineNumbers(output);
};

const cropPadAndNumber = (text, options) => {
	let output = "";

	const lines = text.split("\n");
	const maxDigitWidth = String(
		lines.length + (options.lineNumberStart - 1)
	).length;
	// Tabs needed to contain digits (so we dont break code tab indentation)
	const tabsNeeded = Math.ceil(maxDigitWidth / options.consoleTabWidth);
	const maxTabSpace = tabsNeeded * options.consoleTabWidth;
	const longestLineLength = findLongestLine(text, options);

	options.$maxTabSpace = maxTabSpace;
	options.$maxDigitWidth = maxDigitWidth;

	for (const [i, line] of lines.entries()) {
		const lineNumber = i + 1;
		if (lineNumber < options.firstLine || lineNumber > options.lastLine) {
			continue;
		}

		const lineNo = lineNumberPad(lineNumber, options);

		const tabCount = (line.match(/\t/g) || []).length;
		const tabAdjust = tabCount * options.consoleTabWidth;

		const plain = stripAnsi(line).replace(/\t/g, "");
		const linePad = String().padEnd(
			longestLineLength - plain.length - tabAdjust,
			" "
		);
		const runLengthLine = line + options.colors.trailingSpace(linePad);

		let lineOutput;

		if (options.tabsToSpaces === false) {
			lineOutput = lineNo + runLengthLine;
		}

		if (typeof options.tabsToSpaces === "number") {
			lineOutput = lineNo + padLine(runLengthLine, options.codePad);
		}

		output += lineOutput + "\n";
	}

	return options.colors.base(output);
};

const decorate = (ansiString, options) => {
	if (options.highlight) {
		ansiString = ansiMark(ansiString, options.highlight);
	}

	ansiString = cropPadAndNumber(ansiString, options);
	return ansiString;
};

const nameifyArrowFn = (fn, options) => {
	if (
		Reflect.has(fn, "prototype") &&
		Reflect.has(fn.prototype, "constructor")
	) {
		return "";
	}

	return `${options.arrowKeyword} ${fn.name} = `;
};

const procOptions = (options_ = {}) => {
	let options = {
		lineNumbers: true,
		lang: "javascript",
		lineNumberPad: 0,
		lineNumberStart: 1,
		start: 1,
		end: Number.POSITIVE_INFINITY,
		highlight: false,
		stripIndent: true,
		codePad: 1,
		colors: darkPalette,
		tabsToSpaces: 4,
		consoleTabWidth: 8,
		arrowKeyword: "const",
	};

	options = merge(options, options_);

	options.$indent = {
		spaces: typeof options.tabsToSpaces === "number",
		tabs:
			typeof options.tabsToSpaces === "boolean" &&
			options.tabsToSpaces === false,
		size:
			typeof options.tabsToSpaces === "number" ? options.tabsToSpaces : 1,
	};

	return options;
};

const chromafi = (value, options) => {
	options = procOptions(options);

	if (typeof value === "function") {
		value = nameifyArrowFn(value, options) + String(value);
		const indentStart = true;
		value = syntaxHlString("javascript", value, options, indentStart);
		value = stripIndent(value);
		value = decorate(value, options);
		return value;
	}

	if (typeof value === "string") {
		value = syntaxHlString(options.lang, value, options);
		value = decorate(value, options);
		return value;
	}

	if (typeof value === "object") {
		value = syntaxHlJson(value, options);
		value = decorate(value, options);
		return value;
	}

	throw new Error(
		"🦅  Chromafi: You must pass a function, string or object."
	);
};

export default chromafi;
